type closure = { entry : Id.l; actual_fv : Id.t list }
type t = (* Ѵμ (caml2html: closure_t) *)
  | Unit
  | Int of int
  | Float of float
  | Neg of Id.t
  | Add of Id.t * Id.t
  | Sub of Id.t * Id.t
  | FNeg of Id.t
  | FAdd of Id.t * Id.t
  | FSub of Id.t * Id.t
  | FMul of Id.t * Id.t
  | FDiv of Id.t * Id.t
  | IfEq of Id.t * Id.t * t * t
  | IfLE of Id.t * Id.t * t * t
  | Let of (Id.t * Type.t) * t * t
  | Var of Id.t
  | MakeCls of (Id.t * Type.t) * closure * t
  | AppCls of Id.t * Id.t list
  | AppDir of Id.l * Id.t list
  | Tuple of Id.t list
  | LetTuple of (Id.t * Type.t) list * Id.t * t
  | Get of Id.t * Id.t
  | Put of Id.t * Id.t * Id.t
  | ExtArray of Id.l

type fundef = { name : Id.l * Type.t;
		args : (Id.t * Type.t) list;
		formal_fv : (Id.t * Type.t) list;
		body : t }
type prog = Prog of fundef list * t

let rec fv = function
  | Unit | Int(_) | Float(_) | ExtArray(_) -> S.empty
  | Add(x, y) | Sub(x, y) | FAdd(x, y) | FSub(x, y) | FMul(x, y) | FDiv(x, y) | Get(x, y) -> S.of_list [x; y]
  | Neg(x) | FNeg(x) -> S.singleton x
  | IfEq(x, y, e1, e2)| IfLE(x, y, e1, e2) -> S.add x (S.add y (S.union (fv e1) (fv e2)))
  | Let((x, t), e1, e2) -> S.union (fv e1) (S.remove x (fv e2))
  | Var(x) -> S.singleton x
  | MakeCls((x, t), { entry = l; actual_fv = ys }, e) -> S.remove x (S.union (S.of_list ys) (fv e))
  | AppCls(x, ys) -> S.of_list (x :: ys)
  | AppDir(_, xs) | Tuple(xs) -> S.of_list xs
  | LetTuple(xts, y, e) -> S.add y (S.diff (fv e) (S.of_list (List.map fst xts)))
  | Put(x, y, z) -> S.of_list [x; y; z]

let toplevel : fundef list ref = ref []

let rec g env known = function (* Ѵ롼 (caml2html: closure_g) *)
  | KNormal.Unit -> Unit
  | KNormal.Int(i) -> Int(i)
  | KNormal.Float(d) -> Float(d)
  | KNormal.Neg(x) -> Neg(x)
  | KNormal.Add(x, y) -> Add(x, y)
  | KNormal.Sub(x, y) -> Sub(x, y)
  | KNormal.FNeg(x) -> FNeg(x)
  | KNormal.FAdd(x, y) -> FAdd(x, y)
  | KNormal.FSub(x, y) -> FSub(x, y)
  | KNormal.FMul(x, y) -> FMul(x, y)
  | KNormal.FDiv(x, y) -> FDiv(x, y)
  | KNormal.IfEq(x, y, e1, e2) -> IfEq(x, y, g env known e1, g env known e2)
  | KNormal.IfLE(x, y, e1, e2) -> IfLE(x, y, g env known e1, g env known e2)
  | KNormal.Let((x, t), e1, e2) -> Let((x, t), g env known e1, g (M.add x t env) known e2)
  | KNormal.Var(x) -> Var(x)
  | KNormal.LetRec({ KNormal.name = (x, t); KNormal.args = yts; KNormal.body = e1 }, e2) -> (* ؿξ (caml2html: closure_letrec) *)
      (* ؿlet rec x y1 ... yn = e1 in e2ξϡ
	 x˼ͳѿʤ(closure𤵤direct˸ƤӽФ)
	 Ȳꤷknownɲäe1򥯥ѴƤߤ *)
(*      let _ = Printf.printf "%s : %s\n" x (Type.t_to_string t) in *)      
      let toplevel_backup = !toplevel in
      let env' = M.add x t env in
      let known' = S.add x known in
      let e1' = g (M.add_list yts env') known' e1 in
      (* ˼ͳѿʤäѴe1'ǧ *)
      (* : e1'xȤѿȤƽиclosureɬ!
         (thanks to nuevo-namasute and azounoman; test/cls-bug2.ml) *)
      let zs = S.diff (fv e1') (S.of_list (List.map fst yts)) in
      let known', e1' =
	if S.is_empty zs then known', e1' else
	(* ܤä(toplevel)ᤷơѴľ *)
	(Format.eprintf "free variable(s) %s found in function %s@." (Id.pp_list (S.elements zs)) x;
	 Format.eprintf "function %s cannot be directly applied in fact@." x;
	 toplevel := toplevel_backup;
	 let e1' = g (M.add_list yts env') known e1 in
	 known, e1') in
      let zs = S.elements (S.diff (fv e1') (S.add x (S.of_list (List.map fst yts)))) in (* ͳѿΥꥹ *)
      let zts = List.map (fun z -> (z, M.find z env')) zs in (* Ǽͳѿzη˰envɬ *)
      toplevel := { name = (Id.L(x), t); args = yts; formal_fv = zts; body = e1' } :: !toplevel; (* ȥåץ٥ؿɲ *)
      let e2' = g env' known' e2 in
(*      let _ = Printf.printf "%s : %s\n" x (Type.t_to_string t) in *)
      if S.mem x (fv e2') then (* xѿȤe2'˽и뤫 *)
	MakeCls((x, t), { entry = Id.L(x); actual_fv = zs }, e2') (* иƤʤ *)
      else
	(Format.eprintf "eliminating closure(s) %s@." x;
	 e2') (* иʤMakeCls *)
  | KNormal.App(x, ys) when S.mem x known -> (* ؿŬѤξ (caml2html: closure_app) *)
      Format.eprintf "directly applying %s@." x;
      AppDir(Id.L(x), ys)
  | KNormal.App(f, xs) -> AppCls(f, xs)
  | KNormal.Tuple(xs) -> Tuple(xs)
  | KNormal.LetTuple(xts, y, e) -> LetTuple(xts, y, g (M.add_list xts env) known e)
  | KNormal.Get(x, y) -> Get(x, y)
  | KNormal.Put(x, y, z) -> Put(x, y, z)
  | KNormal.ExtArray(x) -> ExtArray(Id.L(x))
  | KNormal.ExtFunApp(x, ys) -> AppDir(Id.L("min_caml_" ^ x), ys)

let f e =
  toplevel := [];
  let e' = g M.empty S.empty e in
  Prog(List.rev !toplevel, e')


(* print *)
let xt2str (x, t) = "(" ^ x ^ " : " ^ (Type.t_to_string t) ^ ")" (* x *)
let xts2str (x, ts) = "(" ^ x ^ " : " ^ (Type.ts_to_string ts) ^ ")" (* x *)

let print_cls { entry = Id.L l; actual_fv = fvs } = match fvs with
  | [] -> print_string ("[" ^ l ^ "]")
  | a :: rest -> 
      print_string ("[" ^ l ^ ", [" ^ a);
      List.iter (fun x -> print_string (", " ^ x)) rest; 
      print_string "]]"  

let rec print = function 
  | Unit -> print_string "()"
  | Int i -> print_int i
  | Float f -> print_float f
  | Neg x | FNeg x -> print_string ("-" ^ x)
  | Add (x, y) -> print_string (x ^ " + " ^ y)
  | Sub (x, y) -> print_string (x ^ " - " ^ y)
  | FAdd (x, y) -> print_string (x ^ " +. " ^ y)
  | FSub (x, y) -> print_string (x ^ " -. " ^ y)
  | FMul (x, y) -> print_string (x ^ " *. " ^ y)
  | FDiv (x, y) -> print_string (x ^ " /. " ^ y)
  | IfEq (x, y, e1, e2) -> 
      print_string ("if " ^ x ^ " = " ^ y ^ "\n  then ");
      print e1; print_string "\n  else "; print e2
  | IfLE (x, y, e1, e2) ->
      print_string ("if " ^ x ^ " <= " ^ y ^ "\n  then ");
      print e1; print_string "\n  else "; print e2
  | Let (xt, e1, e2) ->
      print_string ("let " ^ (xt2str xt) ^ " = "); print e1;
      print_string " in\n  "; print e2
  | Var x -> print_string x
  | MakeCls (xt, closure, e) -> 
      print_string ("let " ^ (xt2str xt) ^ " = ");
      print_cls closure; 
      print_string " in\n  "; print e
  | AppCls (x, ys) -> 
      print_string ("[" ^ x ^ " ");
      (match ys with
	 | [] -> ()
	 | a :: rest -> 
	     print_string a; 
	     List.iter (fun x -> print_string (" " ^ x)) rest);
      print_string "]"
  | AppDir (Id.L l, ys) -> 
      print_string ("(" ^ l ^ " ");
      (match ys with
	 | [] -> ()
	 | a :: rest -> 
	     print_string a; 
	     List.iter (fun x -> print_string (" " ^ x)) rest);
      print_string ")"
  | Tuple [] | LetTuple ([], _, _) -> assert false 
  | Tuple (x :: rest) ->
      print_string ("(" ^ x);
      List.iter (fun x -> print_string (", " ^ x)) rest;
      print_string ")"
  | LetTuple ((xt :: rest), y, e) ->
      print_string ("(" ^ (xt2str xt));
      List.iter (fun xt -> print_string (", " ^ (xt2str xt))) rest;
      print_string (") = " ^ y ^ " in\n  "); print e
  | Get (x, y) -> Printf.printf "%s.(%s)" x y
  | Put (x, y, z) -> Printf.printf "%s.(%s) <- %s" x y z
  | ExtArray (Id.L l) -> print_string l
      
let print_def { name = (Id.L x, _); args = ys; formal_fv = fvs; body = e } =
  print_string ("let rec " ^ x);
  (match fvs with
     | [] -> ()
     | (a, _) :: rest -> 
	 print_string ("[" ^ a);
	 List.iter (fun (x, _) -> print_string (", " ^ x)) rest;
	 print_string "] ";
	 List.iter (fun (x, _) -> print_string (" " ^ x)) rest);
  print_string " = \n  ";
  print e; print_newline ()

let print_prog (Prog (defs, e)) = 
  print_string "{\n";
  List.iter (fun def -> print_def def) defs;
  print_string "}\n";
  print e
  
