(* PowerPC assembly with a few virtual instructions *)

type id_or_imm = V of Id.t | C of int
type t = (* ̿ *)
  | Ans of exp
  | Let of (Id.t * Type.t) * exp * t
and exp = (* İĤ̿б뼰 *)
  | Nop
  | Li of int
  | FLi of Id.l
  | SetL of Id.l
  | Mr of Id.t
  | Neg of Id.t
  | Add of Id.t * id_or_imm
  | Sub of Id.t * id_or_imm
  | Slw of Id.t * id_or_imm
  | Lwz of Id.t * id_or_imm
  | Stw of Id.t * Id.t * id_or_imm
  | FMr of Id.t 
  | FNeg of Id.t
  | FAdd of Id.t * Id.t
  | FSub of Id.t * Id.t
  | FMul of Id.t * Id.t
  | FDiv of Id.t * Id.t
  | Lfd of Id.t * id_or_imm
  | Stfd of Id.t * Id.t * id_or_imm
  | Comment of string
  (* virtual instructions *)
  | IfEq of Id.t * id_or_imm * t * t
  | IfLE of Id.t * id_or_imm * t * t
  | IfGE of Id.t * id_or_imm * t * t
  | IfFEq of Id.t * Id.t * t * t
  | IfFLE of Id.t * Id.t * t * t
  (* closure address, integer arguments, and float arguments *)
  | CallCls of Id.t * Id.t list * Id.t list
  | CallDir of Id.l * Id.t list * Id.t list
  | Save of Id.t * Id.t (* 쥸ѿͤ򥹥åѿ¸ *)
  | Restore of Id.t (* åѿͤ *)
type fundef =
    { name : Id.l; args : Id.t list; fargs : Id.t list; body : t; ret : Type.t }
(* ץ = ươ֥ + ȥåץ٥ؿ + ᥤμ *)
type prog = Prog of (Id.l * float) list * fundef list * t

(* shorthand of Let for float *)
(* fletd : Id.t * exp * t -> t *)
let fletd (x, e1, e2) = Let ((x, Type.Float), e1, e2)
(* shorthand of Let for unit *)
(* seq : exp * t -> t *)
let seq (e1, e2) = Let ((Id.gentmp Type.Unit, Type.Unit), e1, e2)

let regs = [| "%r5"; "%r6"; "%r7"; "%r8"; "%r9"; "%r10"; "%r11"; 
  "%r12"; "%r13"; "%r14"; "%r15"; "%r16"; "%r17"; "%r18"; "%r19"; 
  "%r20"; "%r21"; "%r22"; "%r23"; "%r24"; "%r25"; "%r26"; "%r27"; 
  "%r28"; "%r29"; "%r30" |]
let fregs = Array.init 32 (fun i -> Printf.sprintf "%%f%d" i)
let allregs = Array.to_list regs
let allfregs = Array.to_list fregs
let reg_cl = regs.(Array.length regs - 1) (* closure address *)
let reg_sw = regs.(Array.length regs - 2) (* temporary for swap *) 
let reg_fsw = fregs.(Array.length fregs - 1) (* temporary for swap *)
let reg_hp = "%r4"   (* heap pointer *)
let reg_sp = "%r3"   (* stack pointer *)
let reg_tmp = "%r31" (* temporary register *)

(* is_reg : Id.t -> bool *)
let is_reg x = x.[0] = '%'

(* remove_and_uniq : S.t -> Id.t list -> Id.t list *)
let rec remove_and_uniq xs = function 
  | [] -> []
  | x :: ys when S.mem x xs -> remove_and_uniq xs ys
  | x :: ys -> x :: remove_and_uniq (S.add x xs) ys

(* free variable in the order of use (for spilling) *)
(* fv_id_or_imm : id_or_imm -> Id.t list *)
let fv_id_or_imm = function V (x) -> [x] | _ -> []
(* fv_exp : t -> S.t list *)
let rec fv_exp = function
  | Nop | Li (_) | FLi (_) | SetL (_) | Comment (_) | Restore (_) -> []
  | Mr (x) | Neg (x) | FMr (x) | FNeg (x) | Save (x, _) -> [x] 
  | Add (x, y') | Sub (x, y') | Slw (x, y') | Lfd (x, y') | Lwz (x, y') -> 
      x :: fv_id_or_imm y'
  | FAdd (x, y) | FSub (x, y) | FMul (x, y) | FDiv (x, y) -> [x; y]
  | Stw (x, y, z') | Stfd (x, y, z') -> x :: y :: fv_id_or_imm z'
  | IfEq (x, y', e1, e2) | IfLE (x, y', e1, e2) | IfGE (x, y', e1, e2) -> 
      x :: fv_id_or_imm y' @ remove_and_uniq S.empty (fv e1 @ fv e2)
  | IfFEq (x, y, e1, e2) | IfFLE (x, y, e1, e2) ->
      x :: y :: remove_and_uniq S.empty (fv e1 @ fv e2)
  | CallCls (x, ys, zs) -> x :: ys @ zs 
  | CallDir (_, ys, zs) -> ys @ zs 
and fv = function 
  | Ans (exp) -> fv_exp exp
  | Let ((x, t), exp, e) ->
      fv_exp exp @ remove_and_uniq (S.singleton x) (fv e)

(* fv : t -> Id.t list *)
let fv e = remove_and_uniq S.empty (fv e)

(* concat : t -> Id.t * Type.t -> t -> t *)
let rec concat e1 xt e2 = match e1 with
  | Ans (exp) -> Let (xt, exp, e2)
  | Let (yt, exp, e1') -> Let (yt, exp, concat e1' xt e2)

(* align : int -> int *)
let align i = if i mod 8 = 0 then i else i + 4


let rec print = function
  | Ans e -> print_string "ans "; print' e
  | Let ((x, _), e, t) -> print_string ("let " ^ x ^ " = "); print' e; print t
and print' = function 
  | Nop -> print_string "nop\n"
  | Li i -> print_string "li "; print_int i; print_newline ()
  | FLi (Id.L x) -> print_string ("fli " ^ x ^ "\n")
  | SetL (Id.L x) -> print_string ("SetL " ^ x ^ "\n")
  | Mr x -> print_string ("mr " ^ x ^ "\n")
  | Neg x -> print_string ("neg " ^ x ^ "\n")
  | Add (x, V y) -> print_string (x ^ " + " ^ y ^ "\n")
  | Add (x, C n) -> print_string (x ^ " + "); print_int n; print_newline ()
  | Sub (x, V y) -> print_string (x ^ " - " ^ y ^ "\n")
  | Sub (x, C n) -> print_string (x ^ " - "); print_int n; print_newline ()
  | Slw (x, V y) -> print_string ("slw " ^ x ^ " " ^ y ^ "\n")
  | Slw (x, C n) -> print_string ("slw " ^ x ^ " "); print_int n; print_newline ()
  | Lwz (x, V y) -> print_string ("lwz " ^ x ^ " " ^ y ^ "\n")
  | Lwz (x, C n) -> print_string ("lwz " ^ x ^ " "); print_int n; print_newline ()
  | Stw (x, y, V z) -> print_string ("stw " ^ x ^ " " ^ y ^ " " ^ z ^ "\n")
  | Stw (x, y, C n) -> print_string ("stw " ^ x ^ " " ^ y ^ " "); print_int n; print_newline ()
  | FMr x -> print_string ("fmr " ^ x ^ "\n")
  | FNeg x -> print_string ("fneg " ^ x ^ "\n")
  | FAdd (x, y) -> print_string (x ^ " +. " ^ y ^ "\n")
  | FSub (x, y) -> print_string (x ^ " -. " ^ y ^ "\n")
  | FMul (x, y) -> print_string (x ^ " *. " ^ y ^ "\n")
  | FDiv (x, y) -> print_string (x ^ " /. " ^ y ^ "\n")
  | Lfd (x, V y) -> print_string ("lfd " ^ x ^ " " ^ y ^ "\n")
  | Lfd (x, C n) -> print_string ("lfd " ^ x ^ " "); print_int n; print_newline ()
  | Stfd (x, y, V z) -> print_string ("stfd " ^ x ^ " " ^ y ^ " " ^ z ^ "\n")
  | Stfd (x, y, C n) -> print_string ("stfd " ^ x ^ " " ^ y ^ " "); print_int n; print_newline ()
  | IfEq (x, V y, t1, t2) -> print_string ("if " ^ x ^ " = " ^ y ^ "\nthen "); print t1; print_string "\nelse "; print t2
  | IfEq (x, C n, t1, t2) -> print_string ("if " ^ x ^ " = " ^ (string_of_int n) ^ "\nthen "); print t1; print_string "\nelse "; print t2
  | IfLE (x, V y, t1, t2) -> print_string ("if " ^ x ^ " <= " ^ y ^ "\nthen "); print t1; print_string "\nelse "; print t2
  | IfLE (x, C n, t1, t2) -> print_string ("if " ^ x ^ " <= " ^ (string_of_int n) ^ "\nthen "); print t1; print_string "\nelse "; print t2
  | IfGE (x, V y, t1, t2) -> print_string ("if " ^ x ^ " >= " ^ y ^ "\nthen "); print t1; print_string "\nelse "; print t2
  | IfGE (x, C n, t1, t2) -> print_string ("if " ^ x ^ " >= " ^ (string_of_int n) ^ "\nthen "); print t1; print_string "\nelse "; print t2
  | IfFEq (x, y, t1, t2) -> print_string ("if " ^ x ^ " = " ^ y ^ "\nthen "); print t1; print_string "\nelse "; print t2
  | IfFLE (x, y, t1, t2) -> print_string ("if " ^ x ^ " <= " ^ y ^ "\nthen "); print t1; print_string "\nelse "; print t2
  | CallCls (x, ints, floats) -> print_string ("callCls " ^ x ^ "\n")
  | CallDir (Id.L x, ints, floats) -> print_string ("callDir " ^ x ^ "\n")
  | Save (x, y) -> print_string ("save " ^ x ^ " " ^ y ^ "\n")
  | Restore x -> print_string ("restore " ^ x ^ "\n")
  | Comment s -> print_string ("(* " ^ s ^ " *)")
 
