(* JFP '98, Hutton & Meijer: Monadic Parsing in Haskell *)

(* DS both for collecting answers and for threading inputs *)

(* item : string -> string *)
let rec item str =
  shift (fun k ->
	   let rec f lst ans = match lst with
	       [] -> ans
	     | first :: rest ->
		 if first + 1 = str + 1
		 then (reset (fun () -> k first)) rest ans
		 else ans
	   in f)
in

(* either : (unit -> 'a) -> (unit -> 'a) -> 'a *)
let rec either p1 p2 =
  shift (fun k ->
	   let rec f lst ans =
	     let ans2 = (reset (fun () -> k (p1 ()))) lst ans
	     in         (reset (fun () -> k (p2 ()))) lst ans2
	   in f)
in

(* many : (unit -> 'a) -> 'a list *)
let rec many p =
  either
    (let rec f unit = [] in f)
    (let rec many1 p = (p () :: many p)
     in let rec f unit = many1 p
     in f)
in

(* parse : (unit -> 'a) -> string list -> ('a * string list) list *)
let rec parse grammar lst =
  reset (fun () -> 
	   let result = grammar () in
	   let rec f lst2 ans = ((result, lst2) :: ans)
	   in f) lst []
in

(* tests *)
let rec test1 unit =
  parse (let rec f unit = item 1 in f) [1; 2]
in
let rec test2 unit =
  parse (let rec f unit = item 1 in f) [2; 3]
in
let rec test3 unit =
  parse (let rec f unit =
	  either (let rec f unit = item 1 in f)
		 (let rec f unit = item 2 in f)
	 in f)
	 [1; 3]
in
let rec test4 unit =
  parse (let rec f unit =
	  either (let rec f unit = item 1 in f)
		 (let rec f unit = item 2 in f)
	 in f)
	[2; 3]
in
let rec test5 unit =
  parse (let rec f unit = many (let rec f unit = item 1 in f)
	 in f)
	[1; 1; 1; 3]
in
let rec test6 unit =
  parse (let rec f unit =
	  let i = many (let rec f unit = item 1 in f) in
	  let c = item 3 in
	  (i, c, item 4)
	 in f)
	[1; 1; 1; 3; 4]

(* (test1 (), test2 (), test3 (), test4 (), test5 (), test6 ()) *)

(* = ([(1, [2])], [], [(1, [3])], [(2, [3])], 
[([1; 1; 1], [3]); ([1; 1], [1; 3]); ([1], [1; 1; 3]); ([], [1; 1; 1; 3])],
[(([1; 1; 1], 3, 4), [])]) *)

in
 
let rec print_il0 lst = match lst with
    [] -> ()
  | a :: rest -> print_int a; print_byte 59; print_byte 32; print_il0 rest 
in

let rec print_il lst = print_byte 91; print_il0 lst; print_byte 93
in

let rec print_i_l0 lst = match lst with
    [] -> ()
  | a :: rest -> 
      let (i, l) = a in 
	(print_byte 40; print_int i; print_byte 44; print_byte 32;
	 print_il l; print_byte 41; 
	 print_byte 59; print_byte 32; print_i_l0 rest)
in

let rec print_i_l lst = print_byte 91; print_i_l0 lst; print_byte 93
in

let rec print_l_l0 lst = match lst with
    [] -> ()
  | a :: rest -> 
      let (l1, l2) = a in
	(print_byte 40; print_il l1; print_byte 44; print_byte 32;
	 print_il l2; print_byte 41;
	 print_byte 59; print_byte 32; print_l_l0 rest)
in

let rec print_l_l lst = print_byte 91; print_l_l0 lst; print_byte 93

in 

let rec print_l_i_i0 lst = match lst with
    [] -> ()
  | a :: rest -> 
      let (p, b) = a in
      let (l, i1, i2) = p in
	(print_byte 40; 
	 print_byte 40; print_il l; print_byte 44; print_byte 32;
	 print_int i1; print_byte 44; print_byte 32;
	 print_int i2; print_byte 44; print_byte 32; print_byte 41;
	 print_byte 44; print_byte 32; print_il b; 
	 print_byte 41; 
	 print_byte 59; print_byte 32; print_l_i_i0 rest)
in 

let rec print_l_i_i lst = print_byte 91; print_l_i_i0 lst; print_byte 93
in

print_i_l (test1 ()); print_newline ();
print_i_l (test2 ()); print_newline ();
print_i_l (test3 ()); print_newline ();
print_i_l (test4 ()); print_newline ();
print_l_l (test5 ()); print_newline ();
print_l_i_i (test6 ()); print_newline () 
