type t =
  | Unit 
  | Bool
  | Int
  | Float
  | Fun of t list * t * t * t
  | Tuple of t list
  | Array of t 
  | List of t
  | Var of t option ref

let rec map_var f = function
  | Fun (t1s, t2, t3, t4) ->
	Fun (List.map (map_var f) t1s, map_var f t2, map_var f t3, map_var f t4)
  | Tuple (ts) -> Tuple (List.map (map_var f) ts)
  | Array (t) -> Array (map_var f t)
  | List (t) -> List (map_var f t)
  | Var ({ contents = None } as r) -> f r
  | Var({ contents = Some(t) }) -> map_var f t
  | t -> t

let subst r r' = map_var (fun r0 -> if r == r0 then r' else Var (r0))

module S : sig
  type 'r t = 'r list
  val empty : 'r t
  val singleton : 'r -> 'r t
  val union : 'r t -> 'r t -> 'r t
  val diff : 'r t -> 'r t -> 'r t
  val fold : ('r -> 'a -> 'a) -> 'r t -> 'a -> 'a
end = struct
  type 'r t = 'r list
  let empty = []
  let singleton a = [a]
  let rec add elt = function
      [] -> [elt]
    | a :: rest -> if a == elt then a :: rest else a :: add elt rest
  let union lst1 lst2 = List.fold_right add lst2 lst1
  let rec del elt = function
      [] -> []
    | a :: rest -> if a == elt then rest else a :: del elt rest
  let diff lst1 lst2 = List.fold_right del lst2 lst1
  let fold = List.fold_right
end

type ts = (* type scheme *)
  t option ref S.t ref * t

let to_ts t = (ref S.empty, t)

let gentyp () = Var(ref None) (* ѿ *)

(* to_string *)

let env = ref ([] : (t option ref * string) list)
let counter = ref (Char.code 'a')
let new_var r =
  let i = !counter in
  let s = "'" ^ Char.escaped (Char.chr i) in
  env := (r, s) :: !env;
  counter := i + 1;
  s

exception EmptyTuple

let rec t_to_string = function
    Unit -> "unit"
  | Bool -> "bool"
  | Int -> "int"
  | Float -> "float"
  | Fun([], t2, t3, t4) ->
      let s2 = t_to_string t2 in 
      let s4 = t_to_string t4 in
	if s2 = s4 
	then "([] -> " ^ (t_to_string t3) ^ ")"
	else "([] / " ^ s2 ^ " -> "  ^ (t_to_string t3) ^ " / " ^ s4 ^ ")"
  | Fun(t1::t1s, t2, t3, t4) ->
      let s2 = t_to_string t2 in
      let s4 = t_to_string t4 in
	if s2 = s4 
	then 
	  "(["
	^ List.fold_left (fun str t -> str ^ ", " ^ t_to_string t)
			 (t_to_string t1) t1s
	^ "] -> " ^ (t_to_string t3) ^ ")"
	else 
	  "(["
	^ List.fold_left (fun str t -> str ^ ", " ^ t_to_string t)
			 (t_to_string t1) t1s
	^ "] / " ^ s2 ^ " -> " ^ (t_to_string t3) ^ " / " ^ s4 ^ ")"
  | Tuple([]) -> raise EmptyTuple
  | Tuple(t :: ts) ->
	"(" ^
	List.fold_left (fun str t -> str ^ " * " ^ t_to_string t)
		       (t_to_string t) ts
	^ ")"
  | Array(t) -> t_to_string t ^ " array"
  | List(t) -> t_to_string t ^ " list"
  | Var(r) -> 
      (match !r with
	 | None -> (try List.assq r !env with Not_found -> new_var r)
	 | Some r' -> t_to_string r')

let rec ts_to_string (rs_ref, t) =
  counter := Char.code 'a';
  let s = S.fold (fun r str -> new_var r ^ ". " ^ str)
		 !rs_ref
		 "\n	" in
  s ^ t_to_string t
